/* stat_alloc.c */

/*******************************************************************/
/** allocation de memoire statique a l'usage des fonctions C      **/
/*******************************************************************/

/* organisation de la memoire statique :

   Tous les blocs comportent un nombre entier de mots (32 bits)
   Les petits blocs (1, 2, 3, 4, ou 5 mots utiles) sont alloues
   par "pages". Les autres sont en vrac.
   Les petits blocs ont un en-tete, les gros ont un en-tete et
   un en-queue.
   Les petits blocs libres sont en liste (une par taille), les gros sont
   en free-list circulaire (doublement chainee.)
*/

/* fonctions de gestion de la memoire statique :
   stat_malloc (x) : alloue un bloc de x octets (renvoie un pointeur).
                     Renvoie 0 si plus de memoire.
   stat_alloc (x) : alloue un bloc de x octets (renvoie un pointeur).
                     Leve l'exception "Out of memory" si plus de memoire.
   stat_free (p) : libere le bloc pointe par p
   stat_resize (p, x) : change la taille du bloc.
     Renvoie la nouvelle adresse du bloc, ou 0 si plus de memoire.
   stat_size (p) : renvoie la taille reelle du bloc. (en octets)
   stat_largest_free_size () : renvoie la taille du plus gros bloc disponible.
                               (en octets)
*/

#include "mlvalues.h"
#include "mem_misc.h"
#include "stat_alloc.h"
#include "fail.h"

typedef unsigned long word;

#define SMALL 5
#define FREE 0
#define ALLOCATED ((unsigned long) 1 << 31)
#define SIZE_MASK (~ALLOCATED)
#define TAG_MASK (ALLOCATED)
#define PAGE_SIZE (4 * 1024)	/* en octets */

/* PREV is only valid for blocks of size >= SMALL
 * This is especially important for blocks of size 1.
 */
#define NEXT(bl) (((word **) (bl)) [0])
#define PREV(bl) (((word **) (bl)) [1])

#define IS_FREE(bl) ((((word *) (bl)) [-1] & TAG_MASK) == FREE)

#define STAT_SIZE(adr) (((word *) (adr)) [-1] & SIZE_MASK)
#define MAKE_STAT_BLOCK(adr, len, tag) \
  { unsigned long l = (len); \
   ((word *) (adr)) [-1] = (l) | (tag); \
   ((word *) (adr)) [l] = (l) | (tag); }

#define BLOCK_BEFORE(bl) ((word *) (bl) - 2 \
			  - (((word *) (bl)) [-2] & SIZE_MASK))
#define BLOCK_AFTER(bl) ((word *) (bl) + STAT_SIZE (bl) + 2)

static word *zone_header;
static word *zone_trailer;
static word *free_list;		/* bloc courant de la free-list */
static word *small_free_lists [SMALL + 1];

/* Appelle le gc majeur pour decaler le tas.
   Quand on retourne vrai, il y a toujours au moins un bloc de
   la taille requise dans la free-list.
*/

static int get_more_memory (request)
     unsigned long request;
{
  VOID *chunk;
  word *block;
  unsigned long length;

#ifdef UNIX_MEMORY
  request = (request + 2) * sizeof (word);
  shift_heap (request + STATIC_ZONE_GROWING, &chunk, &length);
  if (chunk == NULL){
    shift_heap (request, &chunk, &length);
    if (chunk == NULL){
      return 0;
    }
  }
  length = (length / sizeof (word)) - 2;
#else
  extern word * static_end;
  extern word * heap_start;
  extern word * temp_alloc_ptr;
  chunk = static_end;
  if (static_end + request + 2 >= heap_start)
    return 0;
  static_end += request + 2;
  length = request;
  /* The temporary allocation zone is located just after the static zone, and must be
   * erased and shifted. */
  temp_alloc_ptr = static_end;
#endif

  block = ((word *) chunk) + 1;
  Zassert (block == zone_trailer + SMALL + 3);
  block = zone_trailer;
  zone_trailer = block + length + 2;
  MAKE_STAT_BLOCK (block, length, ALLOCATED);
  MAKE_STAT_BLOCK (zone_trailer, SMALL + 1, ALLOCATED);
  stat_free (block);

  return 1;
}

/* Alloue une page de petits blocs de la taille requise.
   Quand on retourne, il y a peut-etre un bloc de cette taille
   dans la free-list correspondante.
*/

static void get_more_small (request)
     unsigned long request;
{
  word *page;
  word *next;
  word *i;

  Zassert (request <= SMALL && request >= 1);
  Zassert (small_free_lists [request] == NULL);

  page = (word *) stat_malloc (PAGE_SIZE);
  next = NULL;
  if (page != NULL) {
    for (i = page + 1; i < page + (PAGE_SIZE / sizeof (word)) - request;
         i += request + 1){
      Zassert ((char *) (i + request + 1) <= (char *) page + PAGE_SIZE);
      i [-1] = request;
      NEXT (i) = next;
      next = i;
    }
  }
  small_free_lists [request] = next;
}

VOID *stat_malloc (size)
     blocksize size;
{
  word *result;
  blocksize w_size = (size + 3) / 4;
  blocksize free_size;

  if (w_size == 0) w_size = 1;

  if (w_size <= SMALL){
				/* Allocation d'un petit bloc */
    if (small_free_lists [w_size] == NULL){
      get_more_small (w_size);	/* On alloue une page. */
    }
    if (small_free_lists [w_size] == NULL){
      return NULL;
    }else{
      result = small_free_lists [w_size];
      small_free_lists [w_size] = NEXT (small_free_lists [w_size]);
      return (VOID *) result;
    }
  }else{
			/* Allocation en free-list, first fit circulaire */
    word *start_point;

    if (free_list != NULL){
      start_point = free_list;
      do{
	if (STAT_SIZE (free_list) >= w_size){
	  goto found;
	}
	free_list = NEXT (free_list);
      }while (free_list != start_point);
    }
				/* Pas de bloc de taille suffisante. */
    if (get_more_memory (w_size)){
      return stat_malloc (size);
    }else{
      return NULL;
    }
  }
 found:
  result = free_list;
  free_size = STAT_SIZE (result);
  Zassert (free_size >= w_size);
  if (free_size > w_size + SMALL + 2){	/* On le coupe en deux. */
    word *new_free = result + w_size + 2;
    blocksize new_size = free_size - w_size - 2;
    word *oldnext, *oldprev;

    MAKE_STAT_BLOCK (new_free, new_size, FREE)
    MAKE_STAT_BLOCK (result, w_size, ALLOCATED);
				/* Chainage. Attention au cas a 1 bloc. */
    oldnext = NEXT (result);
    oldprev = PREV (result);
    NEXT (oldprev) = new_free;
    PREV (oldnext) = new_free;
    NEXT (new_free) = NEXT (result);
    PREV (new_free) = PREV (result);

    free_list = new_free;
    return (VOID *) result;
  }else{
			/* On ne le coupe pas, mais on recolle la free-list. */
    if (NEXT (result) == result){
      free_list = NULL;
    }else{
      NEXT (PREV (result)) = NEXT (result);
      PREV (NEXT (result)) = PREV (result);
      free_list = NEXT (result);
    }
    MAKE_STAT_BLOCK (result, free_size, ALLOCATED);
    return (VOID *) result;
  }
}

VOID *stat_alloc (size)
     blocksize size;
{
  VOID * res;
  res = stat_malloc(size);
  if (res == NULL)
    system_failwith_out_of_memory();
  return res;
}

/* block doit etre un bloc alloue par stat_alloc, ou alors... */

void stat_free (block)
     VOID *block;
{
  blocksize size = STAT_SIZE (block);
  Zassert (size > 0);

  if (size <= SMALL){
    NEXT (block) = small_free_lists [size];
    small_free_lists [size] = (word *) block;
  }else{
    /* Il faut voir si les blocs voisins sont libres, pour les recoller le
       cas echeant. */
    word *bef;
    word *aft;

    Zassert (!IS_FREE (block));

    bef = BLOCK_BEFORE (block);
    aft = BLOCK_AFTER (block);
    if (IS_FREE (bef)){
      if (IS_FREE (aft)){	/* Extraire aft de la liste */
	PREV (NEXT (aft)) = PREV (aft);
	NEXT (PREV (aft)) = NEXT (aft);
	if (free_list == aft) free_list = NEXT (aft);
				/* recoller */
	MAKE_STAT_BLOCK (bef, STAT_SIZE (bef) + size + STAT_SIZE (aft) + 4,
			 FREE);
      }else{
	MAKE_STAT_BLOCK (bef, STAT_SIZE (bef) + size + 2, FREE);
      }
    }else{
      if (IS_FREE (aft)){	/* Attention au cas a 1 bloc. */
	word *oldnext = NEXT (aft);
	word *oldprev = PREV (aft);

	PREV (oldnext) = (word *) block;
	NEXT (oldprev) = (word *) block;
	if (free_list == aft) free_list = (word *) block;
	PREV (block) = PREV (aft);
	NEXT (block) = NEXT (aft);
	MAKE_STAT_BLOCK (block, size + STAT_SIZE (aft) + 2, FREE);
      }else{
	if (free_list == NULL){
	  free_list = (word *) block;
	  PREV (block) = (word *) block;
	  NEXT (block) = (word *) block;
	  MAKE_STAT_BLOCK (block, size, FREE);
	}else{
	  PREV (block) = PREV (free_list);
	  NEXT (PREV (free_list)) = (word *) block;
	  NEXT (block) = free_list;
	  PREV (free_list) = (word *) block;
	  free_list = (word *) block;
	  MAKE_STAT_BLOCK (block, size, FREE);
	}
      }
    }
  }
}

VOID *stat_resize (block, new_size)
     VOID *block;
     blocksize new_size;
{
  blocksize n_size = (new_size + 3) / 4;
  blocksize o_size = STAT_SIZE (block);
  word *new_free;

  if (n_size <= o_size) {
    if (n_size + SMALL + 2 < o_size){
      new_free = (word *) block + n_size + 1;
      MAKE_STAT_BLOCK (block, n_size, ALLOCATED);
      MAKE_STAT_BLOCK (new_free, o_size - n_size - 2, ALLOCATED);
      stat_free (new_free);
    }
    return block;
  }else{
    word *result = (word *) stat_malloc (new_size);
    if (result != NULL){
      bcopy ((char *) block, (char *) result, o_size * sizeof (word));
      stat_free (block);
    }
    return (VOID *) result;
  }
}

blocksize stat_size (block)
     VOID *block;
{
  return STAT_SIZE (block) * sizeof (word);
}

blocksize stat_largest_free_size ()
{
  unsigned long i;

  if (free_list == NULL){
    for (i = SMALL; i > 0; i--){
      if (small_free_lists [i] != NULL){
	return i * sizeof (word);
      }
    }
    return 0;
  }else{
    word *current = free_list;
    unsigned long max = 0;

    do{
      if (STAT_SIZE (current) > max){
	max = STAT_SIZE (current);
      }
      current = NEXT (current);
    }while (current != free_list);
    return max * sizeof (word);
  }
}


/* Pour donner un bout de memoire initial. L'adresse doit etre multiple de 4.

   organisation :

   +---------+--------------------------------------------+----------+
   | header  |                                            | trailer  |
   | de zone |                 zone  statique             | de zone  |
   +---------+--------------------------------------------+----------+

*/

void init_stat_memory (adr, len)
     VOID *adr;
     unsigned long len;
{
  unsigned long i;
  word *free_space;

  len = len / sizeof (word);
  if (len < 3 * (SMALL + 3)){
    fatal_error ("init_stat_memory : initial zone is too small.");
  }

  zone_header = (word *) adr + 1;
  MAKE_STAT_BLOCK (zone_header, SMALL + 1, ALLOCATED);

  free_space = (word *) adr + SMALL + 3 + 1;
  MAKE_STAT_BLOCK (free_space, len - 2 * (SMALL + 3) - 2, ALLOCATED);

  zone_trailer = (word *) adr + len - (SMALL + 3) + 1;
  MAKE_STAT_BLOCK (zone_trailer, SMALL + 1, ALLOCATED);

  stat_free (free_space);

  for (i = 0; i <= SMALL; i++){
    small_free_lists [i] = NULL;
  }
}

